#include "CUDAVolumeTexture.h"

#include <cuda_runtime.h> 
#include <cutil.h>
#include <cutil_math.h>


/// Device Code
texture<char,3,cudaReadModeNormalizedFloat> volumeTex;

__constant__ float c_grey_scale;
__constant__ float c_grey_shift;
__constant__ float4 c_spacings;

static __device__
float d_GetScalarValue(const float4& pos)
{
	return tex3D(volumeTex,pos.x,pos.y,pos.z)*c_grey_scale+c_grey_shift;
}

static __device__
float4 d_GetGradient(const float4& pos)
{
	float4 sample1,sample2;
    sample1.x=tex3D(volumeTex,pos.x-1,pos.y,pos.z);
    sample2.x=tex3D(volumeTex,pos.x+1,pos.y,pos.z);
    sample1.y=tex3D(volumeTex,pos.x,pos.y-1,pos.z);
    sample2.y=tex3D(volumeTex,pos.x,pos.y+1,pos.z);
    sample1.z=tex3D(volumeTex,pos.x,pos.y,pos.z-1);
    sample2.z=tex3D(volumeTex,pos.x,pos.y,pos.z+1);
	sample1.w=0.0f;
	sample2.w=0.0f;
    return (sample2-sample1)*c_grey_scale/c_spacings/2.0f;
}

static __device__
float4 d_GetNormal(const float4& pos)
{
	float4 sample1,sample2;
    sample1.x=tex3D(volumeTex,pos.x-1,pos.y,pos.z);
    sample2.x=tex3D(volumeTex,pos.x+1,pos.y,pos.z);
    sample1.y=tex3D(volumeTex,pos.x,pos.y-1,pos.z);
    sample2.y=tex3D(volumeTex,pos.x,pos.y+1,pos.z);
    sample1.z=tex3D(volumeTex,pos.x,pos.y,pos.z-1);
    sample2.z=tex3D(volumeTex,pos.x,pos.y,pos.z+1);
	sample1.w=0.0f;
	sample2.w=0.0f;
    return normalize((sample2-sample1)/c_spacings);
}

/// Host Code
static bool SingleScalarVolumePrepare(CUDAVolumeTexture *data)
{
	if (!data || !data->m_d_volumeArray) return false;

	volumeTex.normalized = false;  
    volumeTex.filterMode = cudaFilterModeLinear;
    volumeTex.addressMode[0] = cudaAddressModeClamp; 
    volumeTex.addressMode[1] = cudaAddressModeClamp;
	volumeTex.addressMode[2] = cudaAddressModeClamp;

	cudaChannelFormatDesc channelDesc;

	float grey_scale=data->m_grey_scale;
	float grey_shift=data->m_grey_shift;
	cudaArray* d_volumeArray=data->m_d_volumeArray;
	float spacings[4];
	spacings[0]=data->m_Spacings[0];
	spacings[1]=data->m_Spacings[1];
	spacings[2]=data->m_Spacings[2];	
	spacings[3]=1.0f;

	CUDA_SAFE_CALL(cudaGetChannelDesc(&channelDesc, d_volumeArray));
	CUDA_SAFE_CALL(cudaBindTextureToArray(volumeTex, d_volumeArray, channelDesc));

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_grey_scale, &grey_scale, sizeof(float),0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_grey_shift, &grey_shift, sizeof(float),0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_spacings, spacings, sizeof(float)*4,0,cudaMemcpyHostToDevice));
	
	return true;
}

static void SingleScalarVolumeClear()
{
	CUDA_SAFE_CALL(cudaUnbindTexture(volumeTex));
}
