#include <cuda_runtime.h> 
#include <cutil.h>
#include <cutil_math.h>

#include "CUDAVRImage.h"
#include "VRSceneParam.h"

static __device__ 
float4 d_CastRay(const float4& voxel_zero,const float4& voxel_unit,const float4& CameraDirectison,float tnear,float tfar);

__constant__ float c_P2MMatrix[16]; 
__constant__ float c_M2VMatrix[16]; 
__constant__ float c_CroppingBounds[6];

__constant__ int c_imageSize[2];
__constant__ int c_imageOrigin[2];
__constant__ float* c_d_image;
__constant__ float4 c_EyePos;


#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif

static __device__
void GetImageSize(int imageSize[2])
{
	imageSize[0]=c_imageSize[0];
	imageSize[1]=c_imageSize[1];
}

static __device__
float4 mul(const float M[16], const float4 &v)
{
	float4 r;
	r.x = M[0]*v.x+M[4]*v.y+M[8]*v.z+M[12]*v.w;
	r.y = M[1]*v.x+M[5]*v.y+M[9]*v.z+M[13]*v.w;
	r.z = M[2]*v.x+M[6]*v.y+M[10]*v.z+M[14]*v.w;
	r.w = M[3]*v.x+M[7]*v.y+M[11]*v.z+M[15]*v.w;
	return r;
}

static __device__ 
void d_NextRay(const float4& voxel_zero,const float4& CameraDirection, float4& voxel_unit, float& tnear, float& tfar)
{
	voxel_unit=mul(c_M2VMatrix,CameraDirection);

	/// Find Start Point
	float t1,t2;

	t1=(c_CroppingBounds[0]+0.5f-voxel_zero.x)/voxel_unit.x;
	t2=(c_CroppingBounds[1]+0.5f-voxel_zero.x)/voxel_unit.x;

	float tminX=min(t1,t2);
	float tmaxX=max(t1,t2);

	t1=(c_CroppingBounds[2]+0.5f-voxel_zero.y)/voxel_unit.y;
	t2=(c_CroppingBounds[3]+0.5f-voxel_zero.y)/voxel_unit.y;

	float tminY=min(t1,t2);
	float tmaxY=max(t1,t2);

	t1=(c_CroppingBounds[4]+0.5f-voxel_zero.z)/voxel_unit.z;
	t2=(c_CroppingBounds[5]+0.5f-voxel_zero.z)/voxel_unit.z;

	float tminZ=min(t1,t2);
	float tmaxZ=max(t1,t2);

	tnear=max(max(tminX, tminY), tminZ);
	tnear=max(tnear,0.0f);
	tfar=min(min(tmaxX,tmaxY),tmaxZ);
}

static __device__ 
void d_ProduceRay(float pixX,float pixY, float4& voxel_zero,float4& voxel_unit, float4& CameraDirection, float& tnear, float& tfar)
{
	float4 model_pix=mul(c_P2MMatrix,make_float4(pixX,pixY,0.0f,1.0f));
	model_pix*=1.0f/model_pix.w;
	CameraDirection=normalize(model_pix-c_EyePos);

	voxel_zero=mul(c_M2VMatrix,c_EyePos);

	d_NextRay(voxel_zero,CameraDirection, voxel_unit, tnear, tfar);
}

#define iDivUp(a,b) (((a)-1)/(b)+1)

static bool TypicalRayProducerPrepare(CUDAVRImage *image,const VRSceneParam* sceneParam)
{
	if (!image || !sceneParam) return false;

	Matrix4x4 P2MMatrix=sceneParam->m_PixelsToModelMatrix;
	Matrix4x4 M2VMatrix=sceneParam->m_ModelToVolumeMatrix;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_P2MMatrix, P2MMatrix, sizeof(float)*16,0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_M2VMatrix, M2VMatrix, sizeof(float)*16,0,cudaMemcpyHostToDevice));

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_CroppingBounds, sceneParam->m_CroppingBounds, sizeof(float)*6,0,cudaMemcpyHostToDevice));

	int imageSize[2];
	imageSize[0]=image->m_Size[0];
	imageSize[1]=image->m_Size[1];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_imageSize, imageSize, sizeof(int)*2,0,cudaMemcpyHostToDevice));

	int imageOrigin[2];	
	imageOrigin[0]=image->m_Origin[0];
	imageOrigin[1]=image->m_Origin[1];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_imageOrigin, imageOrigin, sizeof(int)*2,0,cudaMemcpyHostToDevice));

	float* d_image=image->m_d_ImageData;
	if (!d_image) 	return false;

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_d_image, &d_image, sizeof(float*),0,cudaMemcpyHostToDevice));

	float EyePos[4];
	EyePos[0]=P2MMatrix[8]/P2MMatrix[11];
	EyePos[1]=P2MMatrix[9]/P2MMatrix[11];
	EyePos[2]=P2MMatrix[10]/P2MMatrix[11];
	EyePos[3]=1.0f;

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_EyePos, EyePos, sizeof(float)*4,0,cudaMemcpyHostToDevice));

	return true;

}
