#include "BasicLightParam.h"

static __device__ float4 d_GetColor(const float4& pos,float sampleDistance);
static __device__ float4 d_GetNormal(const float4& pos);

#include <cuda_runtime.h> 
#include <cutil.h>
#include <cutil_math.h> 

__constant__ float c_sampleDistance;
__constant__ int c_shade;
__constant__ float3 c_ambient;
__constant__ float3 c_diffuse;
__constant__ float3 c_specular;
__constant__ float c_specular_power;
__constant__ float4 c_light_direction;
__constant__ float c_spec_threshold;

__constant__ int c_Illustrative;
__constant__ float3 c_BackgroundColor;

static __device__ 
float4 d_GetShadedColor(const float4& pos,const float4& CameraDirection)
{
	float4 Color=d_GetColor(pos,c_sampleDistance);

	if (Color.w>0.0f && c_shade)
	{
		float4 normal=d_GetNormal(pos);
		float n_dot_c;
		n_dot_c=dot(normal,CameraDirection);

		if (c_Illustrative)
		{
			float rate=min(powf(1.0f-n_dot_c*n_dot_c,c_specular_power*0.5f),1.0f);
			Color.x=Color.x*rate+c_BackgroundColor.x*(1.0f-rate);
			Color.y=Color.y*rate+c_BackgroundColor.y*(1.0f-rate);
			Color.z=Color.z*rate+c_BackgroundColor.z*(1.0f-rate);
		}
		else
		{
			float3 k,b;		
			float n_dot_l;
			float n_dot_h; 
			
			k=c_ambient;
			b=make_float3(0.0f); 
			n_dot_l=dot(normal,c_light_direction);
			
			if ((n_dot_l>0.0f && n_dot_c>0.0f) || (n_dot_l<0.0f && n_dot_c<0.0f))
			{
				n_dot_l=n_dot_l>0?n_dot_l:-n_dot_l;
				k+=c_diffuse*n_dot_l;
				float4 HalfwayVector=normalize(c_light_direction+CameraDirection);
				n_dot_h=dot(normal,HalfwayVector);
				n_dot_h=n_dot_h>0?n_dot_h:-n_dot_h;
				if (n_dot_h > c_spec_threshold)
				{
					b=c_specular*powf(n_dot_h,c_specular_power);
				}
			}

			Color.x=Color.x*k.x+b.x;
			Color.y=Color.y*k.y+b.y;
			Color.z=Color.z*k.z+b.z;
		}
	}

	return Color;
}

static bool FullRayCasterPrepare( bool b_shade, float SampleDistance, const BasicLightParam* param)
{
	if (!param) return false;

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_sampleDistance, &SampleDistance, sizeof(float),0,cudaMemcpyHostToDevice));
	
	int shade=b_shade?1:0;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_shade, &shade, sizeof(int),0,cudaMemcpyHostToDevice));
	
	float lightColor[3];
	lightColor[0]=param->m_LightColor[0];
	lightColor[1]=param->m_LightColor[1];
	lightColor[2]=param->m_LightColor[2];

	float LightIntensity=param->m_LightIntensity;

	lightColor[0]*=LightIntensity;
	lightColor[1]*=LightIntensity;
	lightColor[2]*=LightIntensity;

	float f_ambient=param->m_Ambient;
	float ambient[3];
	ambient[0]=f_ambient*lightColor[0];
	ambient[1]=f_ambient*lightColor[1];
	ambient[2]=f_ambient*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_ambient, ambient, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float f_diffuse=param->m_Diffuse;
	float diffuse[3];
	diffuse[0]=f_diffuse*lightColor[0];
	diffuse[1]=f_diffuse*lightColor[1];
	diffuse[2]=f_diffuse*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_diffuse, diffuse, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float f_specular=param->m_Specular;
	float specular[3];
	specular[0]=f_specular*lightColor[0];
	specular[1]=f_specular*lightColor[1];
	specular[2]=f_specular*lightColor[2];
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_specular, specular, sizeof(float)*3,0,cudaMemcpyHostToDevice));

	float specular_power=param->m_SpecularPower;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_specular_power, &specular_power, sizeof(float),0,cudaMemcpyHostToDevice));

	float spec_threshold=powf(0.001f/max(max(specular[0],specular[1]),specular[2]),1.0f/specular_power);
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_spec_threshold, &spec_threshold, sizeof(float),0,cudaMemcpyHostToDevice));

	float light_direction[4];
	light_direction[0]=param->m_LightDirection[0];
	light_direction[1]=param->m_LightDirection[1];
	light_direction[2]=param->m_LightDirection[2];
	light_direction[3]=0.0f;

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_light_direction, light_direction, sizeof(float)*4,0,cudaMemcpyHostToDevice));

	int illustrative=param->m_Illustrative?1:0;
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_Illustrative, &illustrative, sizeof(int),0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_BackgroundColor, param->m_BackgroundColor, sizeof(float)*3,0,cudaMemcpyHostToDevice));


	return true;
}

